#!/usr/bin/perl

#######################################################
# Script for generating random XOR constraints        #
#                                                     #
# Created by Ashish Sabharwal, Cornell University     #
# Jan 2006                                            #
# Last Modified: Nov 7, 2006                          #
#                                                     #
# Usage example:                                      #
#   1) randxor.pl -n 20 -k 10                         #
#   2) randxor.pl -n 20 -k 10 -f filename -c          #
#   3) randxor.pl -n 20 -k 10 -m 10 -M 20             #
#   4) randxor.pl -n 20 -k 10 -m 10 -l 18 -c          #
#   5) randxor.pl -n 20 -k 10 -m 10 -l 18 -M 20 -f BB #
#                                                     #
# Also prints out the last random number generated    #
# by randxor.pl, for use as a seed when executing     #
# it repeatedly.                                      #
#######################################################

use strict;
use Getopt::Std;
use POSIX;

my $usage = 
    "\nUsage: randxor.pl -n <numvars> -k <numcons> [-c] [-p] [-h]".
    "\n                  [-f <xorfile>] [-s <seed>] [-z <maxnumvars>]".
    "\n                  [-l <avglength>] [-m <minlength>] [-M <maxlength>]".
    "\n".
    "\nAlso prints out the last random number generated internally,".
    "\nfor use as a seed when executing randxor.pl repeatedly".
    "\n".
    "\n  -n numvars     number of variables".
    "\n  -k numcons     number of XOR constraints".
    "\n  -c             generate .cnf in addition to .xor".
    "\n  -p             generate pure .cnf (no aux vars) in addition to .xor".
    "\n  -f xorfile     name of output XOR file; default rand-n-k-l-m-M.xor".
    "\n  -x suffix      suffix for the name of output XOR file".
    "\n  -z maxnumvars  dummy var = maxnumvars+1; default is n".
    "\n  -l avglength   average length of XOR constraints; default floor((m+M)/2)".
    "\n  -m minlength   minimum length of XOR constraints; default 1".
    "\n  -M maxlength   maximum length of XOR constraints; default n".
    "\n  -s seed        random seed; optional; s=0 generates a new seed".
    "\n  -h             help".
    "\n".
    "\nNote: must specify both -n and -k".
    "\nSee the beginning of randxor.pl for some usage examples.".
    "\n\n";


######################
## Global variables ##
######################
my $n;
my $k;
my $l;
my $m;
my $M;
my $s;
my $z;
my @cons;
my $n_cnfvars;
my $n_cnfclauses;
my $n_purecnfclauses;
my $basefilename;
my $xorfilename;
my $cnffilename;
my $purecnffilename;

#############################
# Get current date and time #
#############################
my ($dd, $mm, $yy) = (localtime)[3,4,5];
$yy += 1900;
$mm += 1;
my $fileheader = "c Generated on $mm/$dd/$yy using randxor.pl (created by Ashish Sabharwal)\n";

#############################
## Handle input parameters ##
#############################

our ($opt_n, $opt_k, $opt_c, $opt_p, $opt_f, $opt_x, $opt_z, $opt_l, $opt_m, $opt_M, $opt_s, $opt_h);
getopts('n:k:cpf:x:z:l:m:M:s:h') || die "$usage";

die "\nError: extra arguments $usage" if ($#ARGV != -1);  ## extra unwanted stuff specified

if ($opt_h) {print $usage; exit(0);}

die "\nError: n must be specified\n $usage" if (!$opt_n);
die "\nError: k must be specified\n $usage" if (!$opt_k);

$n = $opt_n;
$k = $opt_k;
$l = $opt_l;
$m = $opt_m;
$M = $opt_M;
$s = $opt_s;
$z = $opt_z;

$m = 1 if ($m == 0);
$M = $n if ($M == 0);
$l = POSIX::floor(($m+$M)/2) if ($l == 0);
$s = time ^ $$ ^ unpack "%L*", `ps axww | gzip` if ($s == 0);
$z = $n if ($z == 0);

($n > 0) || die "n must be positive";
($k > 0) || die "k must be positive";
($m > 0) || die "m must be positive";
($M > 0) || die "M must be positive";
($l >= $m && $l <= $M) || die "l must be between m and M (inclusive)";
($z >= $n) || die "z must be at least n";

if ($opt_f) {
    $basefilename = $opt_f;
    my $suffix = substr($basefilename,-4);
    if ($suffix eq ".xor") {
        $basefilename = substr($basefilename,0,length($basefilename)-4);
    }
}
else {
    $basefilename = "rand-" . substr((100000+$n),1) . 
        "-" . substr((100000+$k),1) .
	"-" . substr((100000+$l),1) . 
        "-" . substr((100000+$m),1) .
	"-" . substr((100000+$M),1);
}
if ($opt_x) {
    $basefilename = $basefilename . "-" . $opt_x;
}

$xorfilename = $basefilename . ".xor";
$cnffilename = $basefilename . ".cnf";
$purecnffilename = $basefilename . "-pure" . ".cnf";



##############################
## Generate XOR constraints ##
##############################

###################
## HACK
print "WARNING: using a hack at the moment to express (a xor b = 0) type constraints\n";
my $kk = $k+1;
my $zz = $z+1;
###################

srand($s);             # feed random seed

$n_cnfvars = $zz;      # number of vars for .cnf
$n_cnfclauses = 0;     # number of clauses for .cnf
$n_purecnfclauses = 0; # number of clauses for pure .cnf

my $p = $l/$n;         # average fraction
my $i = 1;
while ($i <= $k) {
    for my $j (1 .. $n) {
        if (rand(1) < $p) {
            push @{$cons[$i]}, $j;
        }
    }
    my $length = $#{$cons[$i]} + 1;
    if ($length >= $m && $length <= $M) {
        if (rand(1) < 0.5) {   # for (a xor b = 0) constraints
            push @{$cons[$i]}, $zz;
            ++$length;
        }
        if ($opt_c) {  # if .cnf needed, update n_cnfvars and n_cnfclauses
            if ($length >= 2) {
                $n_cnfvars += $length - 2;  # only add auxiliary vars
                $n_cnfclauses += 4*($length - 2) + 2;
            }
            else {
                ++$n_cnfclauses;
            }
        }
        if ($opt_p) {  # if pure .cnf needed, update n_purecnfclauses
            $n_purecnfclauses += (1 << ($length-1));
        }
        ++$i;
    }
    else {
	$#{$cons[$i]} = -1;
    }
}



###########################
## Print XOR constraints ##
###########################

open (XORFILEHANDLE, ">$xorfilename") or die "ERROR: Can't open file \"$xorfilename\": $!";

print "Generating $xorfilename\n";

print XORFILEHANDLE $fileheader,
    "c random XOR constraints\n",
    "c $n variables, $k constraints, dummy variable $zz\n",
    "c average length $l, min length $m, max length $M, random seed $s\n",
    "c .\n",
    "p xor $zz $kk\n";

###################
## HACK
print XORFILEHANDLE "$zz 0\n";
###################

for my $i (1 .. $k) {
    for my $j (0 .. $#{$cons[$i]}) {
        print XORFILEHANDLE $cons[$i][$j], ' ';
    }
    print XORFILEHANDLE "0\n";
}

#print XORFILEHANDLE "0\n";

close XORFILEHANDLE;



#########################################
## Print CNF constraints, if requested ##
#########################################

if ($opt_c) {
    open (CNFFILEHANDLE, ">$cnffilename") or die "ERROR: Can't open file \"$cnffilename\": $!";

    print "Generating $cnffilename\n";

    ##################
    ## HACK
    ++$n_cnfclauses;
    ##################

    print CNFFILEHANDLE $fileheader,
    "c random XOR constraints in CNF form\n",
    "c $n_cnfvars variables (last ", $n_cnfvars - $zz, " auxiliary), dummy variable $zz\n",
    "c $k xor constraints encoded using $n_cnfclauses clauses\n",
    "c average length $l, min length $m, max length $M, random seed $s\n",
    "c .\n",
    "p cnf $n_cnfvars $n_cnfclauses\n";

    ###################
    ## HACK
    print CNFFILEHANDLE "$zz 0\n";
    ###################
    
    my $aux = $zz+1;  # the next auxiliary variable
    for my $i (1 .. $k) {
        my $length = $#{$cons[$i]} + 1;
        if ($length == 1) {
            print CNFFILEHANDLE $cons[$i][0], " 0\n";
        }
        else {
            my $a, $b;
            my $auxprev = $cons[$i][0];
            my $j = 1;
            while ($j < $length-1) {
                $a = $auxprev;
                $b = $cons[$i][$j];
                print CNFFILEHANDLE "$a $b -$aux 0\n";
                print CNFFILEHANDLE "-$a -$b -$aux 0\n";
                print CNFFILEHANDLE "$a -$b $aux 0\n";
                print CNFFILEHANDLE "-$a $b $aux 0\n";
                $auxprev = $aux;
                ++$aux;
                ++$j;
            }
            $a = $auxprev;
            $b = $cons[$i][$length-1];
            print CNFFILEHANDLE "$a $b 0\n";
            print CNFFILEHANDLE "-$a -$b 0\n";
        }
    }

#    print CNFFILEHANDLE "0\n";

    close CNFFILEHANDLE;
}


##############################################
## Print PURE CNF constraints, if requested ##
##############################################

if ($opt_p) {
    open (PURECNFFILEHANDLE, ">$purecnffilename") or die "ERROR: Can't open file \"$purecnffilename\": $!";

    print "Generating $purecnffilename\n";

    ##################
    ## HACK
    ++$n_purecnfclauses;
    ##################

    print PURECNFFILEHANDLE $fileheader,
    "c random XOR constraints in pure CNF form (no auxiliary vars)\n",
    "c $n variables, dummy variable $zz\n",
    "c $k xor constraints encoded using $n_purecnfclauses clauses\n",
    "c average length $l, min length $m, max length $M, random seed $s\n",
    "c .\n",
    "p cnf $zz $n_purecnfclauses\n";

    ###################
    ## HACK
    print PURECNFFILEHANDLE "$zz 0\n";
    ###################
    
    for my $i (1 .. $k) {
        my $length = $#{$cons[$i]} + 1;
        if ($length == 1) {
            print PURECNFFILEHANDLE $cons[$i][0], " 0\n";
        }
        else {
            # use all possible +/- signs for the first $length-1 vars
            # and compute the sign of the last one appropriately
            my @signs;
            for my $j (0 .. $length-2) {
                $signs[$j] = 0;
            }
            $signs[$length-1] = 0;
            for my $count (0 .. (1 << ($length-1))-1) {
                for my $j (0 .. $length-1) {
                    print PURECNFFILEHANDLE '-' if ($signs[$j] == 1);
                    print PURECNFFILEHANDLE $cons[$i][$j], ' ';
                }
                print PURECNFFILEHANDLE "0\n";

                # now compute the next @signs
                my $numsignchange = 0;
                my $position = $length-2;
                while ($signs[$position] == 1 && $position > 0) {
                    $signs[$position] = 0;
                    ++$numsignchange;
                    --$position;
                }
                if ($position < $length-1) {   # otherwise done
                    $signs[$position] = 1;
                    ++$numsignchange;
                    if ($numsignchange % 2 == 1) {
                        $signs[$length-1] = 1 - $signs[$length-1];
                    }
                }
            }
        }
    }

#    print PURECNFFILEHANDLE "0\n";

    close PURECNFFILEHANDLE;
}


# print the last random number generated
print "Last random number generated = ", int(rand(1<<31)), "\n";
